from __future__ import annotations

import functools
from collections import defaultdict
from dataclasses import dataclass
from io import BufferedIOBase
from pathlib import Path
from typing import Sequence, TypeAlias

import numpy as np
import numpy.typing as npt
from scipy.ndimage import generic_filter
from sklearn.cluster import KMeans

import doppy

# ngates, gate points, elevation angle, tuple of sorted azimuth angles
SelectionGroupKeyType: TypeAlias = tuple[int, int, tuple[int, ...]]


@dataclass
class Wind:
    time: npt.NDArray[np.datetime64]
    height: npt.NDArray[np.float64]
    zonal_wind: npt.NDArray[np.float64]
    meridional_wind: npt.NDArray[np.float64]
    vertical_wind: npt.NDArray[np.float64]
    mask: npt.NDArray[np.bool_]
    system_id: str

    @functools.cached_property
    def horizontal_wind_speed(self) -> npt.NDArray[np.float64]:
        return np.sqrt(self.zonal_wind**2 + self.meridional_wind**2)

    @functools.cached_property
    def horizontal_wind_direction(self) -> npt.NDArray[np.float64]:
        direction = np.arctan2(self.zonal_wind, self.meridional_wind)
        direction[direction < 0] += 2 * np.pi
        return np.array(np.degrees(direction), dtype=np.float64)

    @classmethod
    def from_halo_data(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
    ) -> Wind:
        raws = doppy.raw.HaloHpl.from_srcs(data)

        if len(raws) == 0:
            raise doppy.exceptions.NoDataError("HaloHpl data missing")

        raw = (
            doppy.raw.HaloHpl.merge(_select_raws_for_wind(raws))
            .sorted_by_time()
            .non_strictly_increasing_timesteps_removed()
            .nans_removed()
        )
        if len(raw.time) == 0:
            raise doppy.exceptions.NoDataError("No suitable data for the wind product")

        groups = _group_scans_by_azimuth_rotation(raw)
        time_list = []
        elevation_list = []
        wind_list = []
        rmse_list = []

        for group_index in set(groups):
            pick = group_index == groups
            if pick.sum() < 4:
                continue
            time_, elevation_, wind_, rmse_ = _compute_wind(raw[pick])
            time_list.append(time_)
            elevation_list.append(elevation_)
            wind_list.append(wind_[np.newaxis, :, :])
            rmse_list.append(rmse_[np.newaxis, :])
        time = np.array(time_list)
        if len(time) == 0:
            raise doppy.exceptions.NoDataError(
                "Probably something wrong with scan grouping"
            )
        elevation = np.array(elevation_list)
        wind = np.concatenate(wind_list)
        rmse = np.concatenate(rmse_list)
        if not np.allclose(elevation, elevation[0]):
            raise ValueError("Elevation is expected to stay same")
        height = raw.radial_distance * np.sin(np.deg2rad(elevation[0]))
        mask = _compute_mask(wind, rmse)
        return Wind(
            time=time,
            height=height,
            zonal_wind=wind[:, :, 0],
            meridional_wind=wind[:, :, 1],
            vertical_wind=wind[:, :, 2],
            mask=mask,
            system_id=raw.header.system_id,
        )

    @classmethod
    def from_windcube_data(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
    ) -> Wind:
        raws = doppy.raw.WindCube.from_vad_srcs(data)

        if len(raws) == 0:
            raise doppy.exceptions.NoDataError("WindCube data missing")

        raw = (
            doppy.raw.WindCube.merge(raws)
            .sorted_by_time()
            .non_strictly_increasing_timesteps_removed()
            .reindex_scan_indices()
        )
        if len(raw.time) == 0:
            raise doppy.exceptions.NoDataError("No suitable data for the wind product")

        time_list = []
        elevation_list = []
        wind_list = []
        rmse_list = []

        for scan_index in set(raw.scan_index):
            pick = raw.scan_index == scan_index
            if pick.sum() < 4:
                continue
            time_, elevation_, wind_, rmse_ = _compute_wind(raw[pick])
            time_list.append(time_)
            elevation_list.append(elevation_)
            wind_list.append(wind_[np.newaxis, :, :])
            rmse_list.append(rmse_[np.newaxis, :])

        time = np.array(time_list)
        elevation = np.array(elevation_list)
        wind = np.concatenate(wind_list)
        rmse = np.concatenate(rmse_list)
        mask = _compute_mask(wind, rmse)
        if not np.allclose(elevation, elevation[0]):
            raise ValueError("Elevation is expected to stay same")
        height = np.array(raw.height, dtype=np.float64)
        return Wind(
            time=time,
            height=height,
            zonal_wind=wind[:, :, 0],
            meridional_wind=wind[:, :, 1],
            vertical_wind=wind[:, :, 2],
            mask=mask,
            system_id=raw.system_id,
        )

    @classmethod
    def from_wls70_data(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
    ) -> Wind:
        raws = doppy.raw.Wls70.from_srcs(data)

        if len(raws) == 0:
            raise doppy.exceptions.NoDataError("Wls70 data missing")

        raw = (
            doppy.raw.Wls70.merge(raws)
            .sorted_by_time()
            .non_strictly_increasing_timesteps_removed()
        )
        mask = (
            np.isnan(raw.meridional_wind)
            | np.isnan(raw.zonal_wind)
            | np.isnan(raw.vertical_wind)
        )
        return Wind(
            time=raw.time,
            height=raw.altitude,
            zonal_wind=raw.zonal_wind,
            meridional_wind=raw.meridional_wind,
            vertical_wind=raw.vertical_wind,
            mask=mask,
            system_id=raw.system_id,
        )


def _compute_wind(
    raw: doppy.raw.HaloHpl | doppy.raw.WindCube,
) -> tuple[float, float, npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Returns
    -------
    time

    elevation


    wind (range,component):
        Wind components for each range gate.
        Components:
        0: zonal wind
        1: meridional wind
        2: vertical wind

    rmse (range,):
        Root-mean-square error of radial velocity fit for each range gate.
    """
    elevation = np.deg2rad(raw.elevation)
    azimuth = np.deg2rad(raw.azimuth)
    radial_velocity = raw.radial_velocity

    cos_elevation = np.cos(elevation)
    A = np.hstack(
        (
            (np.sin(azimuth) * cos_elevation).reshape(-1, 1),
            (np.cos(azimuth) * cos_elevation).reshape(-1, 1),
            (np.sin(elevation)).reshape(-1, 1),
        )
    )
    A_inv = np.linalg.pinv(A)

    w = A_inv @ radial_velocity
    r_appr = A @ w
    rmse = np.sqrt(np.sum((r_appr - radial_velocity) ** 2, axis=0) / r_appr.shape[0])
    wind = w.T
    time = raw.time[len(raw.time) // 2]
    elevation = np.round(raw.elevation)
    if not np.allclose(elevation, elevation[0]):
        raise ValueError("Elevations in the scan differ")
    return time, elevation[0], wind, rmse


def _compute_mask(
    wind: npt.NDArray[np.float64], rmse: npt.NDArray[np.float64]
) -> npt.NDArray[np.bool_]:
    """
    Parameters
    ----------

    wind (time,range,component)
    intensty (time,range)
    rmse (time,range)
    """

    def neighbour_diff(X: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        mdiff = np.max(np.abs(X - X[len(X) // 2]))
        return np.array(mdiff, dtype=np.float64)

    WIND_NEIGHBOUR_DIFFERENCE = 20
    neighbour_mask = np.any(
        generic_filter(wind, neighbour_diff, size=(1, 3, 1))
        > WIND_NEIGHBOUR_DIFFERENCE,
        axis=2,
    )

    rmse_th = 5
    return np.array((rmse > rmse_th) | neighbour_mask, dtype=np.bool_)


def _group_scans_by_azimuth_rotation(raw: doppy.raw.HaloHpl) -> npt.NDArray[np.int64]:
    max_timedelta_in_scan = np.timedelta64(30, "s")
    if len(raw.time) < 4:
        raise doppy.exceptions.NoDataError(
            "Less than 4 profiles is not sufficient for wind product."
        )
    groups = -1 * np.ones_like(raw.time, dtype=np.int64)

    group = 0
    first_azimuth_of_scan = _wrap_and_round_angle(raw.azimuth[0])
    groups[0] = group
    for i, (time_prev, time, azimuth) in enumerate(
        zip(raw.time[:-1], raw.time[1:], raw.azimuth[1:]), start=1
    ):
        if (
            angle := _wrap_and_round_angle(azimuth)
        ) == first_azimuth_of_scan or time - time_prev > max_timedelta_in_scan:
            group += 1
            first_azimuth_of_scan = angle
        groups[i] = group
    return groups


def _wrap_and_round_angle(a: np.float64) -> int:
    return int(np.round(a)) % 360


def _group_scans(raw: doppy.raw.HaloHpl) -> npt.NDArray[np.int64]:
    if len(raw.time) < 4:
        raise ValueError("Expected at least 4 profiles to compute wind profile")
    if raw.time.dtype != "<M8[us]":
        raise TypeError("time expected to be in numpy datetime[us]")
    time = raw.time.astype(np.float64) * 1e-6
    timediff_in_seconds = np.diff(time)
    kmeans = KMeans(n_clusters=2, n_init="auto").fit(timediff_in_seconds.reshape(-1, 1))
    centers = kmeans.cluster_centers_.flatten()
    scanstep_timediff = centers[np.argmin(centers)]

    if scanstep_timediff < 0.1 or scanstep_timediff > 30:
        raise ValueError(
            "Time difference between profiles in one scan "
            "expected to be between 0.1 and 30 seconds"
        )
    scanstep_timediff_upperbound = 2 * scanstep_timediff
    groups_by_time = -1 * np.ones_like(time, dtype=np.int64)
    groups_by_time[0] = 0
    scan_index = 0
    for i, (t_prev, t) in enumerate(zip(time[:-1], time[1:]), start=1):
        if t - t_prev > scanstep_timediff_upperbound:
            scan_index += 1
        groups_by_time[i] = scan_index

    return _subgroup_scans(raw, groups_by_time)


def _subgroup_scans(
    raw: doppy.raw.HaloHpl, time_groups: npt.NDArray[np.int64]
) -> npt.NDArray[np.int64]:
    """
    Groups scans further based on the azimuth angles
    """
    group = -1 * np.ones_like(raw.time, dtype=np.int64)
    i = -1
    for time_group in set(time_groups):
        i += 1
        (pick,) = np.where(time_group == time_groups)
        raw_group = raw[pick]
        first_azimuth_angle = int(np.round(raw_group.azimuth[0])) % 360
        group[pick[0]] = i
        for j, azi in enumerate(
            (int(np.round(azi)) % 360 for azi in raw_group.azimuth[1:]), start=1
        ):
            if azi == first_azimuth_angle:
                i += 1
            group[pick[j]] = i
    return group


def _select_raws_for_wind(
    raws: Sequence[doppy.raw.HaloHpl],
) -> Sequence[doppy.raw.HaloHpl]:
    if len(raws) == 0:
        raise doppy.exceptions.NoDataError(
            "Cannot select raws for wind from empty list"
        )
    raws_wind = [
        raw
        for raw in raws
        if len(raw.elevation_angles) == 1
        and (el := next(iter(raw.elevation_angles))) < 80
        and el > 25
        and len(raw.azimuth_angles) > 3
    ]
    if len(raws_wind) == 0:
        raise doppy.exceptions.NoDataError(
            "No data suitable for winds: "
            "Multiple elevation angles or "
            "elevation angle >= 80 or "
            "elevation angle <= 25 or "
            "no more than 3 azimuth angles"
        )

    groups: dict[SelectionGroupKeyType, int] = defaultdict(int)

    for raw in raws_wind:
        groups[_selection_key(raw)] += len(raw.time)

    def key_func(key: SelectionGroupKeyType) -> int:
        return groups[key]

    select_tuple = max(groups, key=key_func)

    return [raw for raw in raws_wind if _selection_key(raw) == select_tuple]


def _selection_key(raw: doppy.raw.HaloHpl) -> SelectionGroupKeyType:
    if len(raw.elevation_angles) != 1:
        raise ValueError("Expected only one elevation angle")
    return (
        raw.header.mergable_hash(),
        next(iter(raw.elevation_angles)),
        tuple(sorted(raw.azimuth_angles)),
    )



from typing import TypeVar

T = TypeVar("T")


def merge_all_equal(key: str, lst: list[T]) -> T:
    if len(set(lst)) != 1:
        raise ValueError(f"Cannot merge header key {key} values {lst}")
    return lst[0]



from __future__ import annotations

import io
from dataclasses import dataclass
from datetime import datetime
from io import BufferedIOBase
from pathlib import Path
from typing import Iterable

import numpy as np
import numpy.typing as npt
from numpy import datetime64


@dataclass
class HaloSysParams:
    time: npt.NDArray[datetime64]  # dim: (time, )
    internal_temperature: npt.NDArray[np.float64]  # dim: (time, ), unit: degree Celsius
    internal_relative_humidity: npt.NDArray[np.float64]  # dim: (time, )
    supply_voltage: npt.NDArray[np.float64]  # dim: (time, )
    acquisition_card_temperature: npt.NDArray[np.float64]  # dim: (time, )
    pitch: npt.NDArray[np.float64]  # dim: (time, )
    roll: npt.NDArray[np.float64]  # dim: (time, )

    @classmethod
    def from_src(cls, data: str | Path | bytes | BufferedIOBase) -> HaloSysParams:
        if isinstance(data, str):
            path = Path(data)
            with path.open("rb") as f:
                return _from_src(f)
        elif isinstance(data, Path):
            with data.open("rb") as f:
                return _from_src(f)
        elif isinstance(data, bytes):
            return _from_src(io.BytesIO(data))
        elif isinstance(data, BufferedIOBase):
            return _from_src(data)
        else:
            raise TypeError("Unsupported data type")

    @classmethod
    def merge(cls, raws: Iterable[HaloSysParams]) -> HaloSysParams:
        return cls(
            np.concatenate(tuple(r.time for r in raws)),
            np.concatenate(tuple(r.internal_temperature for r in raws)),
            np.concatenate(tuple(r.internal_relative_humidity for r in raws)),
            np.concatenate(tuple(r.supply_voltage for r in raws)),
            np.concatenate(tuple(r.acquisition_card_temperature for r in raws)),
            np.concatenate(tuple(r.pitch for r in raws)),
            np.concatenate(tuple(r.roll for r in raws)),
        )

    def __getitem__(
        self,
        index: int | slice | list[int] | npt.NDArray[np.int64] | npt.NDArray[np.bool_],
    ) -> HaloSysParams:
        if isinstance(index, (int, slice, list, np.ndarray)):
            return HaloSysParams(
                self.time[index],
                self.internal_temperature[index],
                self.internal_relative_humidity[index],
                self.supply_voltage[index],
                self.acquisition_card_temperature[index],
                self.pitch[index],
                self.roll[index],
            )
        raise TypeError

    def sorted_by_time(self) -> HaloSysParams:
        sort_indices = np.argsort(self.time)
        return self[sort_indices]

    def non_strictly_increasing_timesteps_removed(self) -> HaloSysParams:
        is_increasing = np.insert(np.diff(self.time).astype(int) > 0, 0, True)
        return self[is_increasing]


def _from_src(data: BufferedIOBase) -> HaloSysParams:
    data_bytes = data.read().strip().replace(b",", b".").replace(b"\x00", b"")
    a = data_bytes.strip().split(b"\r\n")
    b = [r.split(b"\t") for r in a]
    arr = np.array(b)
    if arr.shape[1] != 7:
        raise ValueError("Unexpected data fromat")

    def timestr2datetime64_12H(datetime_bytes: bytes) -> np.datetime64:
        return datetime64(
            datetime.strptime(datetime_bytes.decode("utf-8"), "%m/%d/%Y %I:%M:%S %p"),
            "s",
        )

    def timestr2datetime64_24H(datetime_bytes: bytes) -> np.datetime64:
        return datetime64(
            datetime.strptime(datetime_bytes.decode("utf-8"), "%d/%m/%Y %H:%M:%S"), "s"
        )

    arr_time = np.full_like(arr[:, 0], np.datetime64("NaT"), dtype="datetime64[s]")
    for i, time in enumerate(arr[:, 0]):
        try:
            arr_time[i] = timestr2datetime64_12H(time)
        except ValueError:
            arr_time[i] = timestr2datetime64_24H(time)
    arr_others = np.vectorize(float)(arr[:, 1:])
    return HaloSysParams(arr_time, *arr_others.T)



from __future__ import annotations

from dataclasses import dataclass
from io import BufferedIOBase
from pathlib import Path
from typing import Sequence

import numpy as np
import numpy.typing as npt
from netCDF4 import Dataset, num2date
from numpy import datetime64

from doppy.utils import merge_all_equal


@dataclass
class WindCube:
    time: npt.NDArray[datetime64]  # dim: (time, )
    radial_distance: npt.NDArray[np.int64]  # dim: (radial_distance, )
    height: npt.NDArray[np.int64]  # dim: (radial_distance, )
    azimuth: npt.NDArray[np.float64]  # dim: (time, )
    elevation: npt.NDArray[np.float64]  # dim: (time, )
    cnr: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    radial_velocity: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    radial_velocity_confidence: npt.NDArray[np.float64]  # dim: (time, radial_distance)
    scan_index: npt.NDArray[np.int64]
    system_id: str

    @classmethod
    def from_vad_srcs(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
    ) -> list[WindCube]:
        return [WindCube.from_vad_src(src) for src in data]

    @classmethod
    def from_vad_src(cls, data: str | Path | bytes | BufferedIOBase) -> WindCube:
        data_bytes = _src_to_bytes(data)
        nc = Dataset("inmemory.nc", "r", memory=data_bytes)
        return _from_vad_src(nc)

    @classmethod
    def merge(cls, raws: list[WindCube]) -> WindCube:
        return WindCube(
            scan_index=_merge_scan_index([r.scan_index for r in raws]),
            time=np.concatenate([r.time for r in raws]),
            height=_merge_range_vars([r.height for r in raws]),
            radial_distance=_merge_range_vars([r.radial_distance for r in raws]),
            azimuth=np.concatenate([r.azimuth for r in raws]),
            elevation=np.concatenate([r.elevation for r in raws]),
            radial_velocity=np.concatenate([r.radial_velocity for r in raws]),
            radial_velocity_confidence=np.concatenate(
                [r.radial_velocity_confidence for r in raws]
            ),
            cnr=np.concatenate([r.cnr for r in raws]),
            system_id=merge_all_equal("system_id", [r.system_id for r in raws]),
        )

    def __getitem__(
        self,
        index: int
        | slice
        | list[int]
        | npt.NDArray[np.int64]
        | npt.NDArray[np.bool_]
        | tuple[slice, slice],
    ) -> WindCube:
        if isinstance(index, (int, slice, list, np.ndarray)):
            return WindCube(
                time=self.time[index],
                radial_distance=self.radial_distance,
                height=self.height,
                azimuth=self.azimuth[index],
                elevation=self.elevation[index],
                radial_velocity=self.radial_velocity[index],
                radial_velocity_confidence=self.radial_velocity_confidence[index],
                cnr=self.cnr[index],
                scan_index=self.scan_index[index],
                system_id=self.system_id,
            )
        raise TypeError

    def sorted_by_time(self) -> WindCube:
        sort_indices = np.argsort(self.time)
        return self[sort_indices]

    def non_strictly_increasing_timesteps_removed(self) -> WindCube:
        if len(self.time) == 0:
            return self
        mask = np.ones_like(self.time, dtype=np.bool_)
        latest_time = self.time[0]
        for i, t in enumerate(self.time[1:], start=1):
            if t <= latest_time:
                mask[i] = False
            else:
                latest_time = t
        return self[mask]

    def reindex_scan_indices(self) -> WindCube:
        new_indices = np.zeros_like(self.scan_index)
        indexed = set()
        j = 0
        for i in self.scan_index:
            if i in indexed:
                continue
            new_indices[i == self.scan_index] = j
            indexed.add(i)
            j += 1
        self.scan_index = new_indices
        return self


def _merge_range_vars(vars: list[npt.NDArray[np.int64]]) -> npt.NDArray[np.int64]:
    var = np.concatenate([v[np.newaxis, :] for v in vars])
    if not (var[0] == var).all():
        raise ValueError("Cannot merge unequal range variable")
    return np.array(var[0], dtype=np.int64)


def _merge_scan_index(index_list: list[npt.NDArray[np.int64]]) -> npt.NDArray[np.int64]:
    if len(index_list) == 0:
        raise ValueError("cannot merge empty list")

    new_index_list = []
    current_max = index_list[0].max()
    new_index_list.append(index_list[0])

    for index_arr in index_list[1:]:
        new_arr = index_arr + current_max + 1
        new_index_list.append(new_arr)
        current_max += index_arr.max() + 1
    return np.concatenate(new_index_list)


def _src_to_bytes(data: str | Path | bytes | BufferedIOBase) -> bytes:
    if isinstance(data, str):
        path = Path(data)
        with path.open("rb") as f:
            return f.read()
    elif isinstance(data, Path):
        with data.open("rb") as f:
            return f.read()
    elif isinstance(data, bytes):
        return data
    elif isinstance(data, BufferedIOBase):
        return data.read()
    raise TypeError("Unsupported data type")


def _from_vad_src(nc: Dataset) -> WindCube:
    scan_index_list = []
    time_list = []
    cnr_list = []
    radial_wind_speed_list = []
    radial_wind_speed_confidence_list = []
    azimuth_list = []
    elevation_list = []
    range_list = []
    height_list = []

    for i, group in enumerate(
        nc[group] for group in (nc.variables["sweep_group_name"][:])
    ):
        time_list.append(_extract_datetime64_or_raise(group["time"]))
        radial_wind_speed_list.append(
            _extract_float64_or_raise(group["radial_wind_speed"])
        )
        cnr_list.append(_extract_float64_or_raise(group["cnr"]))
        radial_wind_speed_confidence_list.append(
            _extract_float64_or_raise(group["radial_wind_speed_ci"])
        )
        azimuth_list.append(_extract_float64_or_raise(group["azimuth"]))
        elevation_list.append(_extract_float64_or_raise(group["elevation"]))
        range_list.append(_extract_int64_or_raise(group["range"]))
        height_list.append(_extract_int64_or_raise(group["measurement_height"]))
        scan_index_list.append(np.full(group["time"][:].shape, i, dtype=np.int64))

    height = np.concatenate(height_list)
    if not (height == height[0]).all():
        raise ValueError("Unexpected heights")
    radial_distance = np.concatenate(range_list)
    if not (radial_distance == radial_distance[0]).all():
        raise ValueError("Unexpected range")
    return WindCube(
        scan_index=np.concatenate(scan_index_list),
        time=np.concatenate(time_list),
        radial_distance=radial_distance[0],
        height=height[0],
        azimuth=np.concatenate(azimuth_list),
        elevation=np.concatenate(elevation_list),
        radial_velocity=np.concatenate(radial_wind_speed_list),
        radial_velocity_confidence=np.concatenate(radial_wind_speed_confidence_list),
        cnr=np.concatenate(cnr_list),
        system_id=nc.instrument_name,
    )


def _extract_datetime64_or_raise(nc: Dataset) -> npt.NDArray[np.datetime64]:
    match nc.name:
        case "time":
            if nc.dimensions != ("time",):
                raise ValueError
            return np.array(num2date(nc[:], units=nc.units), dtype="datetime64[us]")
        case _:
            raise ValueError(f"Unexpected variable name {nc.name}")


def _extract_float64_or_raise(nc: Dataset) -> npt.NDArray[np.float64]:
    match nc.name:
        case "cnr":
            if nc.dimensions != ("time", "gate_index"):
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "dB":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                raise ValueError
            return np.array(nc[:].data, dtype=np.float64)
        case "radial_wind_speed":
            if nc.dimensions != ("time", "gate_index"):
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "m s-1":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                raise ValueError
            return np.array(nc[:].data, dtype=np.float64)
        case "radial_wind_speed_ci":
            if nc.dimensions != ("time", "gate_index"):
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "percent":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                raise ValueError
            return np.array(nc[:].data, dtype=np.float64)
        case "azimuth" | "elevation":
            if nc.dimensions != ("time",):
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "degrees":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                raise ValueError
            return np.array(nc[:].data, dtype=np.float64)
        case _:
            raise ValueError(f"Unexpected variable name {nc.name}")


def _extract_int64_or_raise(nc: Dataset) -> npt.NDArray[np.int64]:
    match nc.name:
        case "range" | "measurement_height":
            if nc.dimensions != ("time", "gate_index"):
                raise ValueError(f"Unexpected dimensions for {nc.name}")
            if nc.units != "m":
                raise ValueError(f"Unexpected units for {nc.name}")
            if nc[:].mask is not np.bool_(False):
                raise ValueError
            return np.array(nc[:].data, dtype=np.int64)
        case _:
            raise ValueError(f"Unexpected variable name {nc.name}")



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from io import BufferedIOBase
from pathlib import Path
from typing import Any, Sequence

import numpy as np
import numpy.typing as npt
from numpy import datetime64

import doppy
from doppy import exceptions
from doppy.utils import merge_all_equal


@dataclass
class Wls70:
    time: npt.NDArray[datetime64]  # dim: (time, )
    altitude: npt.NDArray[np.float64]  # dim: (altitude, )
    position: npt.NDArray[np.float64]  # dim: (time, )
    temperature: npt.NDArray[np.float64]  # dim: (time, )
    wiper: npt.NDArray[np.bool_]  # dim: (time, )
    cnr: npt.NDArray[np.float64]  # dim: (time, altitude)
    radial_velocity: npt.NDArray[np.float64]  # dim: (time, altitude)
    radial_velocity_deviation: npt.NDArray[np.float64]  # dim: (time, altitude)
    vh: npt.NDArray[np.float64]  # dim: (time, altitude)
    wind_direction: npt.NDArray[np.float64]  # dim: (time, altitude)
    zonal_wind: npt.NDArray[np.float64]  # u := zonal wind?, dim: (time, altitude)
    meridional_wind: npt.NDArray[
        np.float64
    ]  # v := meridional wind?, dim: (time, altitude)
    vertical_wind: npt.NDArray[np.float64]  # w := vertical wind?, dim: (time, altitude)
    system_id: str

    @classmethod
    def from_srcs(
        cls,
        data: Sequence[str]
        | Sequence[Path]
        | Sequence[bytes]
        | Sequence[BufferedIOBase],
    ) -> list[Wls70]:
        if not isinstance(data, (list, tuple)):
            raise TypeError("data should be list or tuple")
        if all(isinstance(src, bytes) for src in data):
            data_bytes = data
        elif all(isinstance(src, str) for src in data):
            data_bytes = []
            for src in data:
                with Path(src).open("rb") as f:
                    data_bytes.append(f.read())
        elif all(isinstance(src, Path) for src in data):
            data_bytes = []
            for src in data:
                with src.open("rb") as f:
                    data_bytes.append(f.read())
        elif all(isinstance(src, BufferedIOBase) for src in data):
            data_bytes = [src.read() for src in data]
        else:
            raise TypeError("Unexpected types in data")
        raws = doppy.rs.raw.wls70.from_bytes_srcs(data_bytes)
        try:
            return [_raw_rs_to_wls70(r) for r in raws]
        except RuntimeError as err:
            raise exceptions.RawParsingError(err) from err

    @classmethod
    def from_src(cls, data: str | Path | bytes | BufferedIOBase) -> Wls70:
        if isinstance(data, str):
            path = Path(data)
            with path.open("rb") as f:
                data_bytes = f.read()
        elif isinstance(data, Path):
            with data.open("rb") as f:
                data_bytes = f.read()
        elif isinstance(data, bytes):
            data_bytes = data
        elif isinstance(data, BufferedIOBase):
            data_bytes = data.read()
        else:
            raise TypeError("Unsupported data type")
        try:
            return _raw_rs_to_wls70(doppy.rs.raw.wls70.from_bytes_src(data_bytes))
        except RuntimeError as err:
            raise exceptions.RawParsingError(err) from err

    def __getitem__(
        self,
        index: int
        | slice
        | list[int]
        | npt.NDArray[np.int64]
        | npt.NDArray[np.bool_]
        | tuple[slice, slice],
    ) -> Wls70:
        if isinstance(index, (int, slice, list, np.ndarray)):
            return Wls70(
                time=self.time[index],
                altitude=self.altitude,
                position=self.position[index],
                temperature=self.temperature[index],
                wiper=self.wiper[index],
                cnr=self.cnr[index],
                radial_velocity=self.radial_velocity[index],
                radial_velocity_deviation=self.radial_velocity_deviation[index],
                vh=self.vh[index],
                wind_direction=self.wind_direction[index],
                zonal_wind=self.zonal_wind[index],
                meridional_wind=self.meridional_wind[index],
                vertical_wind=self.vertical_wind[index],
                system_id=self.system_id,
            )
        raise TypeError

    def sorted_by_time(self) -> Wls70:
        sort_indices = np.argsort(self.time)
        return self[sort_indices]

    @classmethod
    def merge(cls, raws: Sequence[Wls70]) -> Wls70:
        return cls(
            time=np.concatenate(tuple(r.time for r in raws)),
            altitude=raws[0].altitude,
            position=np.concatenate(tuple(r.position for r in raws)),
            temperature=np.concatenate(tuple(r.temperature for r in raws)),
            wiper=np.concatenate(tuple(r.wiper for r in raws)),
            cnr=np.concatenate(tuple(r.cnr for r in raws)),
            radial_velocity=np.concatenate(tuple(r.radial_velocity for r in raws)),
            radial_velocity_deviation=np.concatenate(
                tuple(r.radial_velocity_deviation for r in raws)
            ),
            vh=np.concatenate(tuple(r.vh for r in raws)),
            wind_direction=np.concatenate(tuple(r.wind_direction for r in raws)),
            zonal_wind=np.concatenate(tuple(r.zonal_wind for r in raws)),
            meridional_wind=np.concatenate(tuple(r.meridional_wind for r in raws)),
            vertical_wind=np.concatenate(tuple(r.vertical_wind for r in raws)),
            system_id=merge_all_equal("system_id", [r.system_id for r in raws]),
        )

    def non_strictly_increasing_timesteps_removed(self) -> Wls70:
        if len(self.time) == 0:
            return self
        mask = np.ones_like(self.time, dtype=np.bool_)
        latest_time = self.time[0]
        for i, t in enumerate(self.time[1:], start=1):
            if t <= latest_time:
                mask[i] = False
            else:
                latest_time = t
        return self[mask]


def _raw_rs_to_wls70(
    raw_rs: tuple[dict[str, Any], list[str], npt.NDArray[np.float64]],
) -> Wls70:
    info, cols, data = raw_rs
    altitude = info["altitude"]
    system_id = info["system_id"]
    data = data.reshape(-1, len(cols))
    time_ts = data[:, 0]
    time = np.array([datetime64(datetime.utcfromtimestamp(ts)) for ts in time_ts])

    position = data[:, 1]
    temperature = data[:, 2]
    wiper = np.array(np.isclose(data[:, 3], 1), dtype=np.bool_)
    cnr = data[:, 4::8]
    rws = data[:, 5::8]
    rwsd = data[:, 6::8]
    vh = data[:, 7::8]
    direction = data[:, 8::8]
    u = data[:, 9::8]
    v = data[:, 10::8]
    w = data[:, 11::8]

    return Wls70(
        time=time,
        altitude=altitude,
        position=position,
        temperature=temperature,
        wiper=wiper,
        cnr=cnr,
        radial_velocity=rws,
        radial_velocity_deviation=rwsd,
        vh=vh,
        wind_direction=direction,
        zonal_wind=u,
        meridional_wind=v,
        vertical_wind=w,
        system_id=system_id,
    )
